/*INCLUDES*/
#include <zephyr/kernel.h>
#include <modem/lte_lc.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/drivers/sensor.h>
#include <stdio.h>
#include <zephyr/net/socket.h>
#include <zephyr/net/mqtt.h>
#include "mqtt_connection.h"



/*DEFINES*/
#define SLEEP_TIME_MS 1000 			//1000ms
#define MAIN_SLEEP 5				//sleep time in minutes between measurements

#define HUMID_THRESHOLD 40			//Humidity threshold for evaluation (40 to show function)
#define TEMP_THRESHOLD_COLD 16		//Temperatur threshold for evaluation
#define TEMP_THRESHOLD_HOT 50		//"

/*LEDs*/
#define LED0_NODE DT_ALIAS(led0) 	//DT node identifier for the "led0" alias. RED
#define LED1_NODE DT_ALIAS(led1) 	//DT node identifier for the "led1" alias. GREEN
#define LED2_NODE DT_ALIAS(led2) 	//DT node identifier for the "led2" alias. BLUE

static const struct gpio_dt_spec red_led = GPIO_DT_SPEC_GET(LED0_NODE, gpios);
static const struct gpio_dt_spec green_led = GPIO_DT_SPEC_GET(LED1_NODE, gpios);
static const struct gpio_dt_spec blue_led = GPIO_DT_SPEC_GET(LED2_NODE, gpios);

/*SENSOR*/
const struct device *const dev = DEVICE_DT_GET_ONE(bosch_bme680);
struct sensor_value temp, press, humidity, gas_res;

int sensedQual = 0;					//INT to remember if windows have been opened

/*MQTT*/
/* The mqtt client struct */
static struct mqtt_client client;
/* File descriptor */
static struct pollfd fds;

int dataToSendHumid;				//int for cases which data should be sent
int dataToSendTemp;					//

/*FUNCTIONS*/
int evalSValues();
int printMessageConnection();
int connectedLED();
int airQualLED();
int sensINIT();
int sensValues();
int sendMsg();

/* Semaphore used to block the main thread until the link controller has
 * established an LTE connection.
 */
K_SEM_DEFINE(lte_connected, 0, 1);

/*LTE_HANDLER*/
static void lte_handler(const struct lte_lc_evt *const evt)
{
     switch (evt->type) {
     case LTE_LC_EVT_NW_REG_STATUS:
             if ((evt->nw_reg_status != LTE_LC_NW_REG_REGISTERED_HOME) &&
             (evt->nw_reg_status != LTE_LC_NW_REG_REGISTERED_ROAMING)) {
                     break;
             }

             printk("Connected to: %s network\n",
             evt->nw_reg_status == LTE_LC_NW_REG_REGISTERED_HOME ? "home" : "roaming");

             k_sem_give(&lte_connected);
             break;
     case LTE_LC_EVT_PSM_UPDATE:
     case LTE_LC_EVT_EDRX_UPDATE:
     case LTE_LC_EVT_RRC_UPDATE:			 
     case LTE_LC_EVT_CELL_UPDATE:
     case LTE_LC_EVT_LTE_MODE_UPDATE:
     case LTE_LC_EVT_TAU_PRE_WARNING:
     case LTE_LC_EVT_NEIGHBOR_CELL_MEAS:
     case LTE_LC_EVT_MODEM_SLEEP_EXIT_PRE_WARNING:
     case LTE_LC_EVT_MODEM_SLEEP_EXIT:
     case LTE_LC_EVT_MODEM_SLEEP_ENTER:
             /* Callback events carrying LTE link data */
             break;
     default:
             break;
     }
}

int printMessageConnection(){
	printk("Succsessfully connected to a LTE-Network!\n");
	return 0;
}

int connectedLED(){
	int ret;
	if(!device_is_ready(blue_led.port)){
		return 1;
	}
	ret = gpio_pin_configure_dt(&blue_led, GPIO_OUTPUT_ACTIVE);
	if(ret<0){
		return 1;
	}

	for(int j=0; j<9; j++){
		ret = gpio_pin_toggle_dt(&blue_led);
		if(ret<0){
			return 1;
		}
		k_msleep(500);
	}
	gpio_pin_configure_dt(&blue_led, GPIO_OUTPUT_ACTIVE);
	return 0;
}

int airQualLED(){
	int ret;
	if(!device_is_ready(green_led.port)||!device_is_ready(red_led.port)||!device_is_ready(blue_led.port)){
		return 1;
	}
	if(sensedQual){
		ret = gpio_pin_configure_dt(&red_led, GPIO_OUTPUT_ACTIVE);
		if(ret<0){
			return 1;
		}
	}
	else{
		ret = gpio_pin_configure_dt(&blue_led, GPIO_OUTPUT_INACTIVE);
		if(ret<0){
			return 1;
		}
		ret = gpio_pin_configure_dt(&red_led, GPIO_OUTPUT_INACTIVE);
		if(ret<0){
			return 1;
		}
	}
	ret = gpio_pin_configure_dt(&green_led, GPIO_OUTPUT_ACTIVE);
	if(ret<0){
		return 1;
	}
	for(int j=0; j<9; j++){
		ret = gpio_pin_toggle_dt(&green_led);
		if(ret<0){
			return 1;
		}
		k_msleep(500);
	}
	gpio_pin_configure_dt(&green_led, GPIO_OUTPUT_INACTIVE);	//ALL LEDS 
	gpio_pin_configure_dt(&red_led, GPIO_OUTPUT_INACTIVE);		//SHUT OFF
	gpio_pin_configure_dt(&blue_led, GPIO_OUTPUT_INACTIVE);		//TO SAVE POWER
	return 0;
}

int sensINIT(){	
	printk("Device %p name is %s\n", dev, dev->name);
	return 0;
}

int sensValues(){
	sensor_sample_fetch(dev);
	sensor_channel_get(dev, SENSOR_CHAN_AMBIENT_TEMP, &temp);
	sensor_channel_get(dev, SENSOR_CHAN_PRESS, &press);
	sensor_channel_get(dev, SENSOR_CHAN_HUMIDITY, &humidity);
	sensor_channel_get(dev, SENSOR_CHAN_GAS_RES, &gas_res);

	printk("T: %d.%06d; P: %d.%06d; H: %d.%06d; G: %d.%06d\n",
			temp.val1, temp.val2, press.val1, press.val2,
			humidity.val1, humidity.val2, gas_res.val1,
			gas_res.val2);

	/*Evalueate values to determine if "Stoßlüften" is in order*/
	evalSValues();
	return 0;
}

int evalSValues(){
	if(humidity.val1 >= HUMID_THRESHOLD){
		sensedQual = 1;
		dataToSendHumid = 1;
		printk("Bitte lueften!\n");
	}
	else if(humidity.val1<HUMID_THRESHOLD && sensedQual == 1){
		sensedQual = 0;
		dataToSendHumid = 2;
		printk("Fenster können wieder geschlossen werden.\n");
	}
	else{
		dataToSendHumid = 3;
		printk("Luftfeuchtigkeit gut.\n");
	}
	if(temp.val1<TEMP_THRESHOLD_COLD && sensedQual == 1){
		dataToSendTemp = 1;
		printk("Fenster sollten wieder geschlossen werden.\n");
	}
	else if(temp.val1<TEMP_THRESHOLD_COLD && sensedQual == 0){
		dataToSendTemp = 2;
		printk("Bitte heizen!\n");
	}
	else if(temp.val1>TEMP_THRESHOLD_HOT){
		dataToSendTemp = 3;
		printk("Nehmen Sie mich bitte aus dem Backofen!\n");
	}
	else{
		dataToSendTemp = 4;
		printk("Temperatur gut.\n");
	}
	return 0;
}

int sendMsg(){
	int err;
	switch(dataToSendHumid){
		case 1:
			err = data_publish(&client, MQTT_QOS_1_AT_LEAST_ONCE,
		   			"Bitte lueften!", sizeof("Bitte lueften!")-1);
			if (err) {
				printk("Failed to send message, %d\n", err);
				return 1;	
			}break;
		case 2:
			err = data_publish(&client, MQTT_QOS_1_AT_LEAST_ONCE,
		   			"Bitte die Fenster wieder schließen.", sizeof("Bitte die Fenster wieder schließen.")-1);
			if (err) {
				printk("Failed to send message, %d\n", err);
				return 1;	
			}break;
		case 3:
			err = data_publish(&client, MQTT_QOS_1_AT_LEAST_ONCE,
		   			"Luftfeuchtigkeit gut.", sizeof("Luftfeuchtigkeit gut.")-1);
			if (err) {
				printk("Failed to send message, %d\n", err);
				return 1;	
			}break;
		default: break;
	}
	k_msleep(1000);
	switch(dataToSendTemp){
		case 1:
			err = data_publish(&client, MQTT_QOS_1_AT_LEAST_ONCE,
		   			"Bitte die Fenster wieder schließen.", sizeof("Bitte die Fenster wieder schließen.")-1);
			if (err) {
				printk("Failed to send message, %d\n", err);
				return 1;	
			}break;
		case 2:
			err = data_publish(&client, MQTT_QOS_1_AT_LEAST_ONCE,
		   			"Bitte heizen!", sizeof("Bitte heizen!")-1);
			if (err) {
				printk("Failed to send message, %d\n", err);
				return 1;	
			}break;
		case 3:
			err = data_publish(&client, MQTT_QOS_1_AT_LEAST_ONCE,
		   			"Bitte holen Sie mich aus dem Backofen!", sizeof("Bitte holen Sie mich aus dem Backofen!")-1);
			if (err) {
				printk("Failed to send message, %d\n", err);
				return 1;	
			}break;
		case 4:
			err = data_publish(&client, MQTT_QOS_1_AT_LEAST_ONCE,
		   			"Temperatur gut.", sizeof("Temperatur gut.")-1);
			if (err) {
				printk("Failed to send message, %d\n", err);
				return 1;	
			}break;
		default: break;
	}
	return 0;
}

void main(void)
{
	k_msleep(500);
	gpio_pin_configure_dt(&red_led, GPIO_OUTPUT_ACTIVE);
	/*Initialisierung und Aufbau der Internetverbindung*/
	int err;
	printk("Connecting to LTE network. This may take a few minutes...\n");
	k_msleep(500);
	err = lte_lc_init_and_connect_async(lte_handler);
	if (err) {
			printk("lte_lc_init_and_connect_async, error: %d\n", err);
			return;
	}
	err = lte_lc_psm_req(true); //enabling PSM
	if (err) {
			printk("lte_lc_psm_req, error: %d\n", err);
			return;
	}
	k_sem_take(&lte_connected, K_FOREVER);
	gpio_pin_configure_dt(&red_led, GPIO_OUTPUT_INACTIVE);

	/*Abarbeitung des weiterführenden Programms*/
	printMessageConnection();
	connectedLED();
	sensINIT();	

	err = client_init(&client);
	if (err) {
		printk("Failed to initialize MQTT client: %d\n", err);
		return;
	}

do_connect:
	err = mqtt_connect(&client);
	if (err) {
		printk("Error in mqtt_connect: %d\n", err);
		goto do_connect;
	}
	err = fds_init(&client,&fds);
	if (err) {
		printk("Error in fds_init: %d\n", err);
		return;
	}


	while(1){

		err = poll(&fds, 1, mqtt_keepalive_time_left(&client));
		if (err < 0) {
			printk("Error in poll(): %d\n", errno);
			break;
		}

		err = mqtt_live(&client);
		if ((err != 0) && (err != -EAGAIN)) {
			printk("Error in mqtt_live: %d\n", err);
			break;
		}

		if ((fds.revents & POLLIN) == POLLIN) {
			err = mqtt_input(&client);
			if (err != 0) {
				printk("Error in mqtt_input: %d\n", err);
				break;
			}
		}

		if ((fds.revents & POLLERR) == POLLERR) {
			printk("POLLERR\n");
			break;
		}

		if ((fds.revents & POLLNVAL) == POLLNVAL) {
			printk("POLLNVAL\n");
			break;
		}

		airQualLED();
		sensValues();
		sendMsg();
		k_sleep(K_MINUTES(MAIN_SLEEP));
	}

	printk("Disconnecting MQTT client\n");

	err = mqtt_disconnect(&client);
	if (err) {
		printk("Could not disconnect MQTT client: %d\n", err);
	}
	goto do_connect;
}
